#include "ESPNow.hpp"
#include "LC709203F.h"
#include "NoDataAvailableException.hpp"
#include "dr26.hpp"
#include "f_deep_sleep.hpp"
#include <Arduino.h>
#include <list>

// Execution time:
// FIXME: Boot: needs to be optimised. should be around 300 ms, can be <100ms
// https://github.com/makermoekoe/Picoclick-C3#speed-up-boot-process
// Part				[ms]
// Boot:			 300
// First setup: 	 150 
// Data aquisition:  500	(4 Dendrometers)
// ESPNow setup: 	 220
// Sending 			 580
// TOTAL: 			1450

// Measurement Summary, 2 Dendrometers connected, measuring 4 Dendrometers:
// Duration:                    1.74 [s]
// Current consumption:       141.09 [mA*s]
// Average current:            81.16 [mA]

static const char *TAG = "MAIN";

LC709203F battery_monitor;

ForteDR26 dr26_channel0;
ForteDR26 dr26_channel3;
ForteDR26 dr26_channel1;
ForteDR26 dr26_channel2;

void send_msgs(const std::__cxx11::list<Message> msgs)
{
	for (const Message &msg : msgs) {

		if (msg.send() != ESP_OK) {
			RtcMemory::store_data(msg.getMessageAsMinifiedJsonString());
		}
		unsigned long ts = millis();
		// it takes ~110ms for receiving an acknowledgement by the host in perfect conditions
		uint16_t message_timeout = 2000;
		while (!was_msg_received()) {
			if ((millis() - ts) > message_timeout) {
				RtcMemory::store_data(msg.getMessageAsMinifiedJsonString());
				ESP_LOGE(TAG, "Timeout: Host not available\n");
				break;
			}
		}
		ESP_LOGD(TAG, "Time until acknowledgement: %ld", millis() - ts);
	}
}
// one loop takes ~2200 ms
void setup()
{
	unsigned long ts = millis();
	Serial.begin(115200);
	// Set the GPIO which conrtols the step up to OUTPUT
	gpio_set_direction(GPIO_NUM_32, GPIO_MODE_OUTPUT);

	// blinking led for debug
	// gpio_set_direction(GPIO_NUM_17, GPIO_MODE_OUTPUT);
	// gpio_set_level(GPIO_NUM_17, 1);

	battery_monitor.begin();

	DeepSleep::print_wakeup_reason();
	DeepSleep::bootCount++;
	ESP_LOGD(TAG, "Boot number: %d", DeepSleep::bootCount);
	// battery protection: go to deep sleep for unlimited time when voltage less than 3.2V
	if (battery_monitor.cellVoltage_mV() < 3200) {
		battery_monitor.setPowerMode(LC709203F_POWER_SLEEP);
		esp_deep_sleep_start();
	}

	gpio_set_level(GPIO_NUM_32, 1);
	// delay(100);
	dr26_channel1.setup();
	dr26_channel0.setChannel(0);
	dr26_channel1.setChannel(1);
	dr26_channel2.setChannel(2);
	dr26_channel3.setChannel(3);

	ESP_LOGD(TAG, "Setup took %ld ms", millis() - ts);

	try {
		// FIXME: put me into seperate trys? No data will be sent when 1 exception occurs

		ts = millis();
		auto messages0 = dr26_channel0.buildMessages();
		auto messages1 = dr26_channel1.buildMessages();
		auto messages2 = dr26_channel2.buildMessages();
		auto messages3 = dr26_channel3.buildMessages();
		auto messages4 = battery_monitor.buildMessages();
		// roughly takes 500ms, ~120ms for each adc channel, barely anything for battery monitor
		ESP_LOGD(TAG, "Reading data and building messages took %ld ms", millis() - ts);
		gpio_set_level(GPIO_NUM_32, 0);

		// FIXME: put this outside the try loop?
		ts = millis();
		espnow_setup();
		ESP_LOGD(TAG, "EPSNow setup took %ld ms", millis() - ts);
		ts = millis();
		send_msgs(messages0);
		send_msgs(messages1);
		send_msgs(messages2);
		send_msgs(messages3);
		send_msgs(messages4);
		// roughly takes 3s in ideal conditions
		ESP_LOGD(TAG, "Sending messages took %ld ms", millis() - ts);

	} catch (const NoDataAvailableException &e) {
		std::cerr << e.what() << '\n';
	}
	// just to be safe in case exception happens above
	gpio_set_level(GPIO_NUM_32, 0);
	// keep it in deep sleep
	gpio_hold_en((gpio_num_t)GPIO_NUM_32);

	battery_monitor.setPowerMode(LC709203F_POWER_SLEEP);
	// battery protection: go to deep sleep for unlimited time when voltage less than 3.2V
	if (battery_monitor.cellVoltage_mV() > 3200) {
		DeepSleep::deep_sleep(10 * 60);
	} else {
		esp_deep_sleep_start();
	}
}

void loop() {}